==> DeepMuon-1.12.3.tar.gz: DeepMuon-1.12.3/Tutorial/3.Customize Model.md <==
# Customize Model

## Introduction

Before you create your DL Models, you need to **STRICTLY** follow these regulations:

> 1. Models are saved in python files
>2. All customized models should be stored in folder `models` **unless** the `filepath` of the model is specified in the configuration
> 3. You must obey the basic regulation given by **Pytorch** to write your customized models.
>4. Customized Layers are also saved in the same file of models
> 5. **DO NOT ADD** any executable console scripts into Dataset file, for instance: `print()`,`plt.show()`... all kinds of executable console scripts all forbidden in the file.
>6. **You need to refresh the** `__init__.py` after you add your customized Dataset to make sure `import` works normal.
> 7. Customized Loss Functions are stored in `models/Airloss.py` **or** anywhere else if specify the `filepath`of loss function in the configuration.

## Create Your Model

To create your Model, you need to create a single python file under the path of folder `models` such as `models/examplemodel.py` or anywhere else if you specify the `filepath` of your model in the configuration.

And all Dataset class are stored within your Dataset file `models/examplemodel.py`, Here we give you an example about the customized Dataset class from `models/Pandax4T.py`:

```python
import torch
from torch import nn
class CONV1(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv=nn.Sequential(
            nn.Conv2d(in_channels=1,out_channels=10,kernel_size=(5,5),stride=1),
            nn.BatchNorm2d(10),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=10,out_channels=40,kernel_size=(5,5)),
            nn.BatchNorm2d(40),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=40,out_channels=80,kernel_size=(5,5)),
        )
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(2000, 512),
            nn.BatchNorm1d(512),
            nn.LeakyReLU(),
            nn.Linear(512, 128),
            nn.BatchNorm1d(128),
            nn.LeakyReLU(),
            nn.Linear(128, 2)
        )
    def forward(self, x):   
        x=self.conv(x)
        x=self.flatten(x)
        x = self.linear_relu_stack(x)
        return x
```

## Customized Layer

Customized layers are saved in the same file of your model, here is an example from `models/Hailing.py`:

```pytho
from torch import nn
import torch.nn.functional as F
class HailingDirectNorm(nn.Module):
    def __init__(self) -> None:
        '''
        ## Customized Layer, Normalize the Direction Vector of Hailing Data Derived from _Direct Models
        - Input: [N,3], info: [px,py,pz]
        - Output: [N,3], info: [px,py,pz](Normalized)

        N is the batch size, and the output direction vector is normalized to 1
        '''
        super().__init__()
    def forward(self,x):
        return F.normalize(x)
```

## Customized Loss Function

Customized Loss Functions can be saved in `models/Airloss.py`, here is an example:

```python
import torch
from torch import nn
from torch import Tensor
import numpy as np
class MSALoss(nn.Module):
    def __init__(self,angle_ratio=1):
        """## MSEloss(vec1,vec2)+Angle(vec1,vec2)
        - Args:
            - angle_ratio (int, optional): The ratio to consider the angle loss into total loss. Defaults to 1.
        """
        super().__init__()
        self.angle_ratio=angle_ratio
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        # print(input)
        mseloss=(input-target)**2
        mseloss=torch.sum(mseloss)/(mseloss.shape[0]*mseloss.shape[1])
        # angloss=ang(input=input,target=target)``
        angloss=angn(input=input,target=target)
        return mseloss+self.angle_ratio*angloss
def angn(input,target):
    input=input.detach().cpu().numpy()
    target=target.detach().cpu().numpy()
    res=np.zeros(input.shape[0])
    for i in range(input.shape[0]):
        res[i]=np.dot(input[i],target[i])/(np.sqrt(np.sum(input[i]**2)*np.sum(target[i]**2)))
    res=np.mean(np.arccos(res))
    return torch.from_numpy(np.array(res))
```

==> DeepMuon-1.23.22.tar.gz: DeepMuon-1.23.22/Tutorial/3.Customize Model.md <==
# Customize Model

## Introduction

Before you create your DL Models, you need to **STRICTLY** follow these regulations:

> 1. Models are saved in python files
>2. All customized models should be stored in folder `models` **unless** the `filepath` of the model is specified in the configuration
> 3. You must obey the basic regulation given by **Pytorch** to write your customized models.
>4. Customized Layers are also saved in the same file of models
> 5. **DO NOT ADD** any executable console scripts into Dataset file, for instance: `print()`,`plt.show()`... all kinds of executable console scripts all forbidden in the file.
>6. **You need to refresh the** `__init__.py` after you add your customized Dataset to make sure `import` works normal.
> 7. Customized Loss Functions are stored in `models/Airloss.py` **or** anywhere else if specify the `filepath`of loss function in the configuration.

## Create Your Model

To create your Model, you need to create a single python file under the path of folder `models` such as `models/examplemodel.py` or anywhere else if you specify the `filepath` of your model in the configuration.

And all Dataset class are stored within your Dataset file `models/examplemodel.py`, Here we give you an example about the customized Dataset class from `models/Pandax4T.py`:

```python
import torch
from torch import nn
class CONV1(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv=nn.Sequential(
            nn.Conv2d(in_channels=1,out_channels=10,kernel_size=(5,5),stride=1),
            nn.BatchNorm2d(10),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=10,out_channels=40,kernel_size=(5,5)),
            nn.BatchNorm2d(40),
            nn.LeakyReLU(),
            nn.Conv2d(in_channels=40,out_channels=80,kernel_size=(5,5)),
        )
        self.flatten = nn.Flatten()
        self.linear_relu_stack = nn.Sequential(
            nn.Linear(2000, 512),
            nn.BatchNorm1d(512),
            nn.LeakyReLU(),
            nn.Linear(512, 128),
            nn.BatchNorm1d(128),
            nn.LeakyReLU(),
            nn.Linear(128, 2)
        )
    def forward(self, x):   
        x=self.conv(x)
        x=self.flatten(x)
        x = self.linear_relu_stack(x)
        return x
```

## Customized Layer

Customized layers are saved in the same file of your model, here is an example from `models/Hailing.py`:

```pytho
from torch import nn
import torch.nn.functional as F
class HailingDirectNorm(nn.Module):
    def __init__(self) -> None:
        '''
        ## Customized Layer, Normalize the Direction Vector of Hailing Data Derived from _Direct Models
        - Input: [N,3], info: [px,py,pz]
        - Output: [N,3], info: [px,py,pz](Normalized)

        N is the batch size, and the output direction vector is normalized to 1
        '''
        super().__init__()
    def forward(self,x):
        return F.normalize(x)
```

## Customized Loss Function

Customized Loss Functions can be saved under `loss_fn`, here is an example of `MSALoss` saved in `loss_fn/MSALoss.py`:

```python
import torch
from torch import nn
from torch import Tensor
import numpy as np
class MSALoss(nn.Module):
    def __init__(self,angle_ratio=1):
        """## MSEloss(vec1,vec2)+Angle(vec1,vec2)
        - Args:
            - angle_ratio (int, optional): The ratio to consider the angle loss into total loss. Defaults to 1.
        """
        super().__init__()
        self.angle_ratio=angle_ratio
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        # print(input)
        mseloss=(input-target)**2
        mseloss=torch.sum(mseloss)/(mseloss.shape[0]*mseloss.shape[1])
        # angloss=ang(input=input,target=target)``
        angloss=angn(input=input,target=target)
        return mseloss+self.angle_ratio*angloss
def angn(input,target):
    input=input.detach().cpu().numpy()
    target=target.detach().cpu().numpy()
    res=np.zeros(input.shape[0])
    for i in range(input.shape[0]):
        res[i]=np.dot(input[i],target[i])/(np.sqrt(np.sum(input[i]**2)*np.sum(target[i]**2)))
    res=np.mean(np.arccos(res))
    return torch.from_numpy(np.array(res))
```
